
from .decorators import decoratormethod

class HTTPMethodsMixIn():
    """convenience methods for .route()"""

    @decoratormethod
    def get(self, *args, **kwargs):
        return self.route(*args,  methods=['GET'], **kwargs)

    @decoratormethod
    def post(self, *args, **kwargs):
        return self.route(*args, methods=['POST'], **kwargs)

    @decoratormethod
    def put(self, *args, **kwargs):
        return self.route(*args, methods=['PUT'], **kwargs)

    @decoratormethod
    def delete(self, *args, **kwargs):
        return self.route(*args, methods=['DELETE'], **kwargs)

    @decoratormethod
    def patch(self, *args, **kwargs):
        return self.route(*args, methods=['PATCH'], **kwargs)

    @decoratormethod
    def options(self, *args, **kwargs):
        return self.route(*args, methods=['OPTIONS'], **kwargs)


class RESTFulRoutesMixIn(HTTPMethodsMixIn):
    """convenience methods for .route()

    https://medium.com/@shubhangirajagrawal/the-7-restful-routes-a8e84201f206
    https://gist.github.com/alexpchin/09939db6f81d654af06b
    """

    resource_id = '<int:id>'

    @decoratormethod
    def index(self, endpoint, *paths, **kwargs):
        "display a list of this resource"
        return self.get(endpoint, *paths or [''], **kwargs)

    @decoratormethod
    def new(self, endpoint, *paths, **kwargs):
        "show a form to create this resource"
        for path in paths or ['']:
            self.get(endpoint, f'new{path}', **kwargs)
        return endpoint

    @decoratormethod
    def create(self, endpoint, *paths, **kwargs):
        "add a new resource to database, then redirect"
        return self.post(endpoint, *paths or [''], **kwargs)

    @decoratormethod
    def show(self, endpoint, *paths, **kwargs):
        "show info about one resource"
        for path in paths or ['']:
            self.get(endpoint, f'/{self.resource_id}{path}', **kwargs)
        return endpoint

    @decoratormethod
    def edit(self, endpoint, *paths, **kwargs):
        "show a form to edit one resource"
        for path in paths or ['']:
            self.get(endpoint, f'/{self.resource_id}/edit{path}', **kwargs)
        return endpoint

    @decoratormethod
    def update(self, endpoint, *paths, **kwargs):
        "update a particular resource, then redirect"
        for path in paths or ['']:
             self.route(endpoint, f'/{self.resource_id}{path}', methods=['PATCH', 'PUT'], **kwargs)
        return endpoint

    @decoratormethod
    def destroy(self, endpoint, *paths, **kwargs):
        "delete a particular resource, then redirect"
        for path in paths or ['']:
             self.delete(endpoint, f'/{self.resource_id}{path}', **kwargs)
        return endpoint



from functools import wraps
from typing import Iterable, Callable
from inspect import signature

__all__ = 'decorator', 'decoratormethod', 'before', 'after', 'decorators', 'wraps_signature'


def wraps_signature(wrapped):
    "functools.wraps but preserving the argument signature"
    def wrapper(func):
        funky = wraps(wrapped)(func)
        funky.__signature__ = signature(wrapped)
        return funky
    return wrapper


def decorator(decfunc):

    def on_decorate(*args, **kwargs):
        if args and callable(args[0]):
            # @decfunc
            return decfunc(*args, **kwargs)
        else:
            # @decfunc(x=1)
            def apply(func):
                return decfunc(func, **kwargs)
            return apply
    return on_decorate


def decoratormethod(decfunc):
    #print('method', decfunc)
    def on_decorate(self, *args, **kwargs):
        #print('methodondec', args, kwargs)
        if args and callable(args[0]):
            # @decfunc
            return decfunc(self, *args, **kwargs)
        else:
            # @decfunc(x=1)
            def apply(func):
                return decfunc(self, func, *args, **kwargs)
            return apply
    return on_decorate


@decorator
def before(func):
    "call before the decorated function runs"
    @decorator
    def decorate(func2):
        @wraps_signature(func2)
        def wrap(*args, **kwargs):
            func()
            return func2(*args, **kwargs)
        return wrap
    return decorate

@decorator
def after(func):
    "call after the decorated function runs"
    @decorator
    def decorate(func2):
        @wraps_signature(func2)
        def wrap(*args, **kwargs):
            value = func2(*args, **kwargs)
            func()
            return value
        return wrap
    return decorate


def decorators(*decorators: Iterable[Callable]):
    "bundle many decorators"
    def decorate(func):
        for dec in decorators:
            func = dec(func)
        return func
    return decorate



class decoratorlist(list):
    """
    useful as a class attribute e.g

    self.things = decoratorlist()

    @obj.things
    def my_function():
        pass

    obj.things -> [my_function]

    """

    @decoratormethod
    def __call__(self, func):
        self.append(func)




import json
from dataclasses import dataclass
from dataclasses_json import dataclass_json

import madness

# create marshmallow schema by decorating a class
schema = lambda cls: dataclass_json(dataclass(cls)).schema()

def response(obj, **kwargs):
    return madness.response(
        [json.dumps(obj)],
        mimetype = 'application/json; charset=utf-8',
        **kwargs
    )

def request(*args):
    data = json.loads(madness.request.get_data())
    if args:
        return (data[arg] for arg in args)
    return data



#!/usr/bin/env python3.7

from dataclasses import dataclass

from werkzeug.wrappers import Response as response
from werkzeug.utils import redirect

from .aws import LambdaMixIn
from .cors import CORSMixIn
from .restful import RESTFulRoutesMixIn
from .application import ApplicationMixIn
from .context import request, context

__all__ = ()

@dataclass
class Madness(CORSMixIn, LambdaMixIn, RESTFulRoutesMixIn, ApplicationMixIn):
    """
    """



from inspect import isgenerator, getfullargspec

from werkzeug.local import Local, LocalManager, LocalProxy, LocalStack

from .exceptions import MissingContextVariable

__all__ = 'request', 'context', 'local_manager', 'local', 'Context', 'bind'

local = Local()
local_manager = LocalManager([local])
request = local('request')
context = local('context')

class Context(dict):

    __setattr__ = dict.__setitem__
    __getattr__ = dict.__getitem__
    __delattr__ = dict.__delitem__

    def run(self, func):
        "pass kwargs directly from this context"
        kwargs = {}
        for key in getfullargspec(func).args:
            try:
                value = self[key]
            except KeyError:
                raise MissingContextVariable(key, func)
            else:
                kwargs[key] = value
        return func(**kwargs)



def bind(endpoint, context_decorators):
    """binds the endpoint to the context
    """

    if not context_decorators:
        return endpoint

    def wrapped_endpoint():
        # run context before/after the endpoint
        # context may catch exceptions

        exitstack = []

        try:
            for func in context_decorators:
                gen = context.run(func)
                if isgenerator(gen):
                    try:
                        # context yielded during except handler
                        response = gen.send(None)
                        if response != None:
                            #print(gen, 'except handler returned', response)
                            break
                    except StopIteration:
                        pass
                    else:
                        exitstack.append(gen)
            else:
                # context did not abort the request
                response = context.run(endpoint)

        except Exception as exception:
            # allow context to recover from an error
            current_exception = exception
            response = None
        else:
            current_exception = None

        #print('unstack', repr(current_exception), 'and', response)

        for gen in reversed(exitstack):
            try:
                if current_exception == None:
                    #print('send', response, 'into', gen)
                    new_response = gen.send(response)
                else:
                    #print('throw', repr(current_exception), 'into', gen)
                    new_response = gen.throw(current_exception)
            except StopIteration:
                pass
            except HTTPException as new_response:
                # abort()/raise used to create a new response
                # continue calling exception handlers
                current_exception = new_response
                response = None
            except Exception as exception:
                current_exception = exception
                response = None
            else:
                if new_response != None:
                    #print(gen, 'modified response', repr(response), 'type=', type(response))
                    current_exception = None
                    response = new_response

        if current_exception != None:
            raise current_exception

        return response

    return wrapped_endpoint





from dataclasses import dataclass, field
from typing import Callable, List, Tuple

from more_itertools import collapse
from werkzeug.routing import Rule

from .context import bind



@dataclass
class Route():
    """
    path does not begin with a leading slash
    """
    path: str
    endpoint: Callable
    methods: List[str] = field(default_factory=list)
    context: List = field(default_factory=list)

    @property
    def rule(self) -> Rule:
        return Rule(
            f'/{self.path}',
            endpoint = bind(self.endpoint, self.context_decorators),
            methods = self.methods or None
        )

    @property
    def context_decorators(self) -> Tuple:
        """
        """
        return tuple(collapse(self.context))

    def __repr__(self):
        context = ','.join([func.__qualname__ for func in self.context_decorators])
        if context:
            context = f'-> @({context})'
        methods = ','.join(self.methods)
        if methods:
            methods = f'[{methods}]'
        parts = ' '.join(
            filter(
            bool, (
            self.__class__.__name__,
            f'/{self.path}',
            methods,
            context,
            f'-> {self.endpoint.__qualname__}',
        )))
        return f'<{parts}>'




from functools import wraps
from typing import Callable, Tuple, List, Any

from werkzeug.wrappers import Response

from .decorators import decoratormethod

class CORSMixIn():
    "https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS"

    @decoratormethod
    def route(self, endpoint:Callable, *paths:Tuple[str], origin:Any=[], max_age:int=None, vary:str=None, methods:List[str]=[], headers:List[str]=None, **kwargs) -> Callable:
        """
        origin may be a list of urls or '*'
        headers is a whitelist of request headers
        default is to allow all headers
        """

        if 'OPTIONS' in methods:
            @wraps(endpoint)
            def wrapped_endpoint():
                if request.method == 'OPTIONS':

                    options_headers = {}

                    request_origin = request.environ.get('HTTP_ORIGIN', '*')

                    if request_origin in origin:
                        options_headers['Access-Control-Allow-Origin'] = request_origin
                        options_headers['Access-Control-Allow-Methods'] = ', '.join(methods)
                        if headers != None:
                            options_headers['Access-Control-Allow-Headers'] = ', '.join(headers)

                    if vary:
                        options_headers['Vary'] = vary

                    if max_age != None:
                        options_headers['Access-Control-Max-Age'] = str(max_age)

                    return Response([], headers=options_headers)
                else:
                    return context.run(endpoint)
        else:
            wrapped_endpoint = endpoint

        super().route(wrapped_endpoint, *paths, methods=methods, **kwargs)
        return endpoint




from json import loads

from marshmallow.schema import Schema
from marshmallow.exceptions import ValidationError

from . import json
from .context import request, context
from .decorators import decoratormethod, decoratorlist

def validation_error_to_lambda_error():
    try:
        yield
    except ValidationError as error:
        yield json.response(
            {
                'errorMessage': ', '.join(
                    f'{key}: {" ".join(value)}'
                    for key, value in error.messages.items()
                ),
                'errorType': 'ValidationError',
                'stackTrace': [],
            },
            status = 422
        )


class LambdaMixIn():

    @decoratormethod
    def lambda_handler(self, endpoint, *args, context=[], proxy_integration=False, loads=loads, **kwargs):
        """

        """

        if proxy_integration:
            raise NotImplementedError('proxy integration')

        else:
            context = decoratorlist([context])

            schema = endpoint.__annotations__.get('event')

            @context
            def parse_event():
                global context
                nonlocal loads
                context.event = loads(request.get_data())

            if isinstance(schema, Schema):
                context(validation_error_to_lambda_error)
                @context
                def validate_event(event):
                    "raises ValidationError"
                    nonlocal schema
                    schema.load(event)

            @context
            def encode_lambda_response():
                obj = yield
                yield json.response(obj)

        return self.route(endpoint, *args, context=context, **kwargs)




from dataclasses import dataclass, replace, field
from typing import Any, Callable, Iterable, List, Dict, Tuple

from .decorators import decoratormethod, decoratorlist
from .route import Route


@dataclass
class RoutingMixIn():
    """
    analogous to flask.Flask and flask.Blueprint combined
    """

    _application_changed: bool = True

    context: List[Callable] = field(default_factory=decoratorlist)

    routes: List[Route] = field(default_factory=list)

    def _add_route(self, route):
        self._application_changed = True
        self.routes.append(route)

    def __iter__(self):
        yield from self.routes

    @decoratormethod
    def route(self, endpoint, *paths, methods=[], context=[]):

        for path in paths or [endpoint.__name__]:
            route = Route(
                path = path,
                endpoint = endpoint,
                methods = methods,
                context = [self.context, context],
            )
            self._add_route(route)
        return endpoint

    def extend(self, routes: Iterable[Route], path='', context=True):
        """

        path should be empty or start with a slash

        e.g.
            api.extend(routes, '/v5')
        or
            versions.extend(routes, '/v5')
            api.extend(versions)

        if context is True, this context will wrap the routes

        """
        for route in routes:
            route = replace(route, path = path + route.path)
            if context:
                route = replace(route, context = [self.context, route.context])
            self._add_route(route)



from typing import Dict, Callable
from dataclasses import dataclass, field
from pprint import pprint

from werkzeug.wrappers import Request
from werkzeug.exceptions import HTTPException#, NotFound, Unauthorized, abort
from werkzeug.routing import Map
from werkzeug.serving import run_simple

from .decorators import decoratormethod
from .context import *

from .routing import RoutingMixIn

@dataclass
class ApplicationMixIn(RoutingMixIn):

    _application = None

    errors: Dict[int, Callable] = field(default_factory=dict)

    @decoratormethod
    def error(self, endpoint, status_code=None):
        """register an error handler e.g. 404
        should be used on the top level routes
        """
        self.errors[status_code] = endpoint
        return endpoint

    def get_application(self, debug:bool=False) -> Callable:

        mapper = Map([route.rule for route in self])

        errors = {
            code: bind(endpoint, self.context)
            for code, endpoint in self.errors.items()
        }

        if debug:
            print('-'*10, 'routes', '-'*10)
            for route in self:
                print(route)
            print('-'*28)
            pprint(errors)
            print('-'*28)

        @local_manager.make_middleware
        def wsgi_application(environ, start_response):
            global local, Request
            nonlocal mapper, errors
            local.request = Request(environ)
            local.context = Context(debug=debug)
            adapter = mapper.bind_to_environ(environ)
            try:
                endpoint, kwargs = adapter.match()
            except HTTPException as response:
                try:
                    endpoint = errors[response.code]
                except:
                    try:
                        endpoint = errors[None]
                    except KeyError:
                        return response(environ, start_response)
                response = endpoint()
                return response(environ, start_response)
            else:
                try:
                    # add path parameters to context
                    context.update(kwargs)
                    response = context.run(endpoint)
                    return response(environ, start_response)
                except HTTPException as response:
                    return response(environ, start_response)

        return wsgi_application

    def __call__(self, environ, start_response):
        """the WSGI application"""
        if self._application_changed:
            self._application = self.get_application()
            self._application_changed = False
        return self._application(environ, start_response)

    def run(self, *, port=9090, host='127.0.0.1', debug=True):
        """run development server"""
        run_simple(
            host,
            port,
            self.get_application(debug=debug),
            use_debugger = debug,
            use_reloader = debug
        )
